#include "pbrt.h"
#include "camera.h"
#include "paramset.h"
#include "film.h"
#include "scene.h"

#include "camerasensor.h"
#include "simplesampler.h"


#include <fstream>
#include <algorithm>
#include <vector>
#include <string>

using namespace std;


// Example representation of an autofocus zone.  The struct contains a CameraSensor
// object to hold the values of pixels in this region, as well as a
// SimpleStratifiedSampler which is need to generate samples the eventually get
// passed along to Scene::Li()
class AfZone {
	public:

	  // from the data file
	  float left, right;
	  float top, bottom;

	  CameraSensor* sensor;
	  SimpleStratifiedSampler* sampler;
};

class Lens {
	public:

		float radius, axpos;
		float N, aperture;
		float axposcum, midpoint;
};


class RealisticCamera : public Camera {
public:
   RealisticCamera(const Transform &world2cam,
      float hither, float yon, float sopen,
      float sclose, float filmdistance, float aperture_diameter, 
      const string &specfile,
	  const string &autofocusfile,
      float filmdiag,
	  Film *film);
   ~RealisticCamera();
   float GenerateRay(const Sample &sample, Ray *ray) const;
   void  AutoFocus(Scene* scene);
   void  ParseLens(const string& filename);
   void  ParseAfZones(const string& filename);
   bool RealisticCamera::Intersect(Ray &r, float& tHit, Point& pHit, float radius) const;

private:
   bool  autofocus;
   float FilmDistance, FilmDiag;
   float Aperture;
   vector<AfZone> afZones;
   vector<Lens> lenses;
};


RealisticCamera::RealisticCamera(const Transform &world2cam,
                                 float hither, float yon, 
                                 float sopen, float sclose, 
                                 float filmdistance, float aperture_diameter, 
                                 const string &specfile,
								 const string &autofocusfile,
                                 float filmdiag,
								 Film *f)
                                 : Camera(world2cam, hither, yon, sopen, sclose, f)
{
	FilmDistance = filmdistance;
	FilmDiag = filmdiag;
	Aperture = aperture_diameter;

	// Parse camera lens file
	ParseLens(specfile);

	// If 'autofocusfile' is the empty string, then you should do
	// nothing in any subsequent call to AutoFocus()
	autofocus = false;

	if (autofocusfile.compare("") != 0)  {		
		ParseAfZones(autofocusfile);
		autofocus = true;		
	}
}

// Parse the camera lens file
void RealisticCamera::ParseLens(const string& filename)
{
  ifstream specfile(filename.c_str());
   if (!specfile) {
      fprintf(stderr, "Cannot open file %s\n", filename.c_str());
      exit (-1);
   }

   char line[512];
   while (!specfile.eof()) {
      specfile.getline(line, 512);
      if (line[0] != '\0' && line[0] != '#' &&
         line[0] != ' ' && line[0] != '\t' && line[0] != '\n')
	  {
		  lenses.resize(lenses.size()+1);
		  Lens& lens = lenses[lenses.size()-1];
		  sscanf(line, "%f %f %f %f\n", &lens.radius, &lens.axpos, &lens.N, &lens.aperture);
		  if (lenses.size() == 1) {
			  lens.axposcum = lens.axpos;
			  lens.midpoint = -lens.radius;
		  } 
		  else if (lenses[lenses.size()-2].radius == 0.0) { 
			  if (lens.radius < 0) {
				  lens.axposcum = lenses[lenses.size()-2].axposcum + lens.axpos;
				  lens.midpoint = -lens.radius - lens.axposcum; 
			  } 
			  else {
				  lens.axposcum = lenses[lenses.size()-2].axposcum + lens.axpos;
				  lens.midpoint = -lens.radius - lens.axposcum;
			  }
		  }
		  else if (lens.radius < 0) {
			  lens.midpoint = -lens.radius - lenses[lenses.size()-2].axposcum;
			  lens.axposcum = lenses[lenses.size()-2].axposcum + lens.axpos;
		  }
		  else {
			  lens.midpoint = -lens.radius - lenses[lenses.size()-2].axposcum;
			  lens.axposcum = lenses[lenses.size()-2].axposcum + lens.axpos;
		  }
	  }
   }
	printf("Read in %d lines from %s\n", lenses.size(), filename.c_str());
}

// parses the AF zone file
void RealisticCamera::ParseAfZones(const string& filename)
{
  ifstream specfile(filename.c_str());
   if (!specfile) {
      fprintf(stderr, "Cannot open file %s\n", filename.c_str());
      exit (-1);
   }

   char line[512];

   while (!specfile.eof()) {
      specfile.getline(line, 512);
      if (line[0] != '\0' && line[0] != '#' &&
         line[0] != ' ' && line[0] != '\t' && line[0] != '\n') 
      {
		afZones.resize(afZones.size()+1);
		AfZone& zone = afZones[afZones.size()-1];
		sscanf(line, "%f %f %f %f\n", &zone.left, &zone.right, &zone.top, &zone.bottom);
      }
   }

	printf("Read in %d AF zones from %s\n", afZones.size(), filename.c_str());
}


RealisticCamera::~RealisticCamera() 
{

}
 

float RealisticCamera::GenerateRay(const Sample &sample, Ray *ray) const 
{
  // YOUR CODE HERE -- make that ray!
  
  // use sample->imageX and sample->imageY to get raster-space coordinates
  // of the sample point on the film.
  // use sample->lensU and sample->lensV to get a sample position on the lens
  
  // GenerateRay() should return the weight of the generated ray

	// Determine z-position of the film plane
	float filmplane = -(FilmDistance + lenses[lenses.size()-1].axposcum);
	float DiskRadius = lenses[lenses.size()-1].aperture / 2.f;

	// Generate point on the film plane
	Point Pras(sample.imageX, sample.imageY, filmplane); 

	// Scale film plane to match the film diameter filmdiag
	Pras.x = -(Pras.x * ((FilmDiag * sqrt(2.f) / 2.f) / film->xResolution) - 0.5 * (FilmDiag * sqrt(2.f) / 2.f));
	Pras.y = Pras.y * ((FilmDiag * sqrt(2.f) / 2.f) / film->yResolution) - 0.5 * (FilmDiag * sqrt(2.f) / 2.f);

	// Generate a sample on an uniareal disk with the size of the first lensradius
	float lensU, lensV;
	ConcentricSampleDisk(sample.lensU, sample.lensV,
		                     &lensU, &lensV);
	lensU *= DiskRadius;
	lensV *= DiskRadius;

	// Set the initial origin and direction of the ray
	ray->o = Pras;
	Point pDisk(lensU, lensV, filmplane + FilmDistance);
	ray->d = pDisk - ray->o;
	ray->d = Normalize(ray->d);

	// Trace ray through lens system
	Point pHit;
	float tHit;
	Vector I, T; 
	Vector N;
	Ray ray_obj;
	Ray ray_aperture;
	float Theta, R, Adisk;

	for (int i=1; i<=lenses.size(); i++) {
		// Set helper rays 
		ray_obj.o = ray->o;
		ray_obj.d = ray->d;
		ray_aperture.o = ray->o;
		ray_aperture.d = ray->d;

		// Retrieve lens element parameters
		float radius = lenses[lenses.size()-i].radius;
		float aperture = lenses[lenses.size()-i].aperture;
		float n1 = lenses[lenses.size()-i].N;
		float n2;
		if (i == lenses.size())
			n2 = 1.f;
		else {
			n2 = lenses[lenses.size()-i-1].N; 
			if (n2 == 0.f)
				n2 = 1.f;
		}

		// Check if ray hits the camera aperture
		if (radius == 0.0) {
			// Intersect ray with aperture element (disk)
			//ray_aperture.o.z = ray_aperture.o.z + (lenses[lenses.size()-i].axposcum + (lenses[lenses.size()-i].axpos / 2.f));
			ray_aperture.o.z = ray_aperture.o.z + (lenses[lenses.size()-i-1].axposcum);
			float thit = -ray_aperture.o.z / ray_aperture.d.z;
			Point phit = ray_aperture(thit);
			float dist2 = phit.x * phit.x + phit.y * phit.y;
			if (dist2 > (Aperture / 2.f) * (Aperture / 2.f)) {
				return 0.f;
			}
			continue;
		}


		// Transform ray_obj to object space
		ray_obj.o.z = ray_obj.o.z - lenses[lenses.size()-i].midpoint;

		// Compute quadratic sphere coefficients
		float A = ray_obj.d.x*ray_obj.d.x + ray_obj.d.y*ray_obj.d.y +
				  ray_obj.d.z*ray_obj.d.z;
		float B = 2.f * (ray_obj.d.x*ray_obj.o.x + ray_obj.d.y*ray_obj.o.y +
					   ray_obj.d.z*ray_obj.o.z);
		float C = ray_obj.o.x*ray_obj.o.x + ray_obj.o.y*ray_obj.o.y +
				  ray_obj.o.z*ray_obj.o.z - radius*radius;
		// Solve quadratic equation for _t_ values 
		float t0, t1; 
		if (!Quadratic(A, B, C, &t0, &t1))
			return 0.f;
		// Compute intersection distance along ray_obj
		if (t0 < 0.f)
			tHit = t1;
		else if (radius > 0.f)
			tHit = t1;
		else tHit = t0;
		// Compute sphere hit position
		pHit = ray_obj(tHit);

		// Check if ray pierces through lens element's aperture
		if (i < lenses.size()) {
			float dist2 = pHit.x * pHit.x + pHit.y * pHit.y;
			if (dist2 > (aperture / 2.f) * (aperture / 2.f)) {
				return 0.f;
			}
		}

		// Compute normal at pHit
		Point PCenter = (0.f,0.f,0.f);
		N = pHit - PCenter;
		if (N.z > 0.f)
			N = -N;
			//N = -1.f * N;
		N = Normalize(N);
		float mu, gamma; 
		I = Vector(ray_obj.d);
		I = Normalize(I);
		mu = n1 / n2; 
		float theta1 = acos(AbsDot(I,N));
		if (i == 1) {
			Theta = theta1;
			Vector r = pHit - ray_obj.o;
			R = r.Length();
		}
		float test = 1 - (pow(mu,2) * (1 - pow((AbsDot(I,N)),2)));
		if (test < 0){
			return 0.f;
		}
		float theta2 = asin(mu * sin(theta1));
		gamma = mu * cos(theta1) - cos(theta2);
		T = mu * I + gamma * N;
		// Transform ray to next lens element position
		pHit.z = pHit.z + lenses[lenses.size()-i].midpoint;
		ray->o = pHit;
		ray->d = Normalize(T);
	}

	// Transform ray into world space
	CameraToWorld(*ray, ray);

	// Calculate and return pixel-weight
	Adisk = DiskRadius * DiskRadius * M_PI;
	float weight = (cos (Theta) * cos (Theta)) / (R * R / Adisk);
	return weight; 
}

void  RealisticCamera::AutoFocus(Scene* scene)
{

	if (!autofocus)
		return;

	printf("Camera has %d autofocus zones.\n", afZones.size()); 


	// setup AF zones
	for (size_t i=0; i<afZones.size(); i++) {
		
		// Compute the number of pixels in this AF zone
		int zone_xres = abs((afZones[i].right - afZones[i].left) * film->xResolution);
		int zone_yres = abs((afZones[i].top - afZones[i].bottom) * film->yResolution);

		// Allocate the zone's CameraSensor object to store the image
		CameraSensor* sensor = new CameraSensor(zone_xres, zone_yres);
		// Allocate the zone's SimpleStratifiedSampler object
		// jitterSensor -- whether or not to jitter spatial samples on the sensor
		// jitterNonSensor -- whether or not to jitter all other samples
		bool b_jitterSensor = 0;
		bool b_jitterNonSensor = 0;
		int SSsamples = 16;
		SimpleStratifiedSampler* sampler = new SimpleStratifiedSampler(zone_xres, zone_yres, SSsamples, b_jitterSensor, b_jitterNonSensor);
	}


	// The following code is stolen from the Scene::Render() function and modified to
	// fit the needs of CS348B Assignment 3.  It should serve as an example of how
	// to render the part of the scene that lies within the an autofocus zone.  I do 
	// it for the 1st AF zone here:

	Sample *sample = new Sample(scene->surfaceIntegrator,
								scene->volumeIntegrator,
								scene);				

	afZones[0].sampler->reset(); 
	afZones[0].sensor->ResetImage();

	while (afZones[0].sampler->GetNextSample(sample)) {

		int pixelX = Floor2Int(sample->imageX);
		int pixelY = Floor2Int(sample->imageY);

		// HINT: YOUR CODE HERE: Before the call to GenerateRay() sample->imageX
		// and sample->imageY need to per perturbed so that they give the index of pixels on
		// real film plane, not the pixels within the current AF zone.  You need to modify
		// sample->imageX and sample->imageY here to perform this change

		float q = sample->imageX;
		float w = sample->imageY;
		int r = 1;

		// Find camera ray for _sample_
		RayDifferential ray;
		float rayWeight = GenerateRay(*sample, &ray);
		// Generate ray differentials for camera ray
		++(sample->imageX);
		GenerateRay(*sample, &ray.rx);
		--(sample->imageX);
		++(sample->imageY);
		GenerateRay(*sample, &ray.ry);
		ray.hasDifferentials = true;
		--(sample->imageY);
		// Evaluate radiance along camera ray
		float alpha;
		Spectrum Ls = 0.f;
		if (rayWeight > 0.f)
			Ls = rayWeight * scene->Li(ray, sample, &alpha);
		// Issue warning if unexpected radiance value returned
		if (Ls.IsNaN()) {
			Error("Not-a-number radiance value returned "
				  "for image sample.  Setting to black.");
			Ls = Spectrum(0.f);
		}
		else if (Ls.y() < -1e-5) {
			Error("Negative luminance value, %g, returned "
				  "for image sample.  Setting to black.", Ls.y());
			Ls = Spectrum(0.f);
		}
		else if (isinf(Ls.y())) {
			Error("Infinite luminance value returned "
				  "for image sample.  Setting to black.");
			Ls = Spectrum(0.f);
		}
		// Add sample contribution to image
		afZones[0].sensor->AddSample(pixelX, pixelY, Ls, alpha);
		// Free BSDF memory from computing image sample value
		BSDF::FreeAll();						
	}

	delete sample;


	float* rgb = afZones[0].sensor->ComputeImageRGB();

	// YOUR CODE HERE! The rbg contents of the image for this zone
	// are now stored in the array 'rgb'.  You can now do whatever
	// processing you wish





	// autofocusing complete. Clean up
	for (size_t i=0; i<afZones.size(); i++) {
		delete afZones[i].sensor;
		delete afZones[i].sampler;
	}
}


extern "C" DLLEXPORT Camera *CreateCamera(const ParamSet &params,
                                          const Transform &world2cam, Film *film) 
{
   // Extract common camera parameters from \use{ParamSet}
   float hither = params.FindOneFloat("hither", -1);
   float yon = params.FindOneFloat("yon", -1);
   float shutteropen = params.FindOneFloat("shutteropen", -1);
   float shutterclose = params.FindOneFloat("shutterclose", -1);

   // Realistic camera-specific parameters
   string specfile = params.FindOneString("specfile", "");
   float filmdistance = params.FindOneFloat("filmdistance", 70.0); // about 70 mm default to film
   float fstop = params.FindOneFloat("aperture_diameter", 1.0);	
   float filmdiag = params.FindOneFloat("filmdiag", 35.0);
   string autofocusfile = params.FindOneString("af_zones", "");

   assert(hither != -1 && yon != -1 && shutteropen != -1 &&
      shutterclose != -1 && filmdistance!= -1);
   if (specfile == "") {
       Severe( "No lens spec file supplied!\n" );
   }
   return new RealisticCamera(world2cam, hither, yon,
      shutteropen, shutterclose, filmdistance, fstop, 
      specfile, autofocusfile, filmdiag, film);
}
